#include <stdio.h>
#include <string.h>
#include "judge.h"

#define MAXN 10000
#define MAXM 1000000
#define MAXD 100
#define MAXH (MAXM+MAXM/3)

unsigned n, m;
unsigned deg[MAXN+1];

struct edge {
  unsigned int u, v, d;
} hash[MAXH];

static unsigned hashf(unsigned u, unsigned v)
{
  return (259309*u + 17321*v) % MAXH;
}

static unsigned edge_find(unsigned u, unsigned v)
{
  unsigned h = hashf(u, v);
  while (hash[h].d)
    {
      if (hash[h].u == u && hash[h].v == v)
	return hash[h].d;
      h = (h+1) % MAXH;
    }
  return 0;
}

static void edge_add(unsigned u, unsigned v, unsigned d)
{
  unsigned h = hashf(u, v);
  while (hash[h].d)
    h = (h+1) % MAXH;
  hash[h] = (struct edge) { .u = u, .v = v, .d = d };
}

static void read_graph(struct tokenizer *t)
{
  t->flags |= TF_REPORT_LINES;
  n = get_uint(t);
  m = get_uint(t);
  if (n < 4 || m < 4 || n > MAXN || m > MAXM)
    tok_err(t, "Interni chyba: M nebo N mimo rozsah");
  get_nl(t);

  for (unsigned i=0; i<m; i++)
    {
      unsigned u, v, d;
      u = get_uint(t);
      v = get_uint(t);
      if (u < 1 || u > n || v < 1 || v > n)
	tok_err(t, "Interni chyba: hrana %d,%d mimo rozsah", u, v);
      d = get_uint(t);
      if (d < 1 || d > 256)
        tok_err(t, "Interni chyba: vaha hrany mimo rozsah");
      if (deg[u]++ > MAXD)
	tok_err(t, "Interni chyba: vrchol %d ma moc velky stupen", u);
      if (deg[v]++ > MAXD)
	tok_err(t, "Interni chyba: vrchol %d ma moc velky stupen", v);
      if (edge_find(u, v))
	tok_err(t, "Interni chyba: hrana %d,%d je nasobna", u, v);
      edge_add(u, v, d);
      edge_add(v, u, d);
      get_nl(t);
    }
  if (get_token(t))
    tok_err(t, "Interni chyba: smeti na konci vstupu");
}

static unsigned get_v(struct tokenizer *t)
{
  unsigned v = get_int(t);
  if (v < 1 || v > n)
    tok_err(t, "Vrchol mimo rozsah: %u", v);
  return v;
}

static unsigned cost(struct tokenizer *t, unsigned u, unsigned v)
{
  unsigned c = edge_find(u, v);
  if (!c)
    tok_err(t, "Hrana %d,%d neexistuje", u, v);
  return c;
}

int main(int argc, char **argv)
{
  if (argc != 4)
    {
      fprintf(stderr, "Usage: judge <in> <out> <correct>\n");
      return 2;
    }

  struct tokenizer tin, tout, tcorr;
  tok_init(&tin, sopen_read(argv[1]));
  tok_init(&tout, sopen_read(argv[2]));
  tok_init(&tcorr, sopen_read(argv[3]));

  read_graph(&tin);

  // Tady narozdil od vstupu na koncich radku netrvame
  char *tok = get_token(&tout);
  if (!tok)
    tok_err(&tout, "Prazdny vystup");
  if (!strcmp(tok, "NEEXISTUJE"))
    {
      char *tok2 = get_token(&tcorr);
      if (strcmp(tok, tok2))
	tok_err(&tout, "Chybne hlasi neexistenci");
      if (get_token(&tout))
	tok_err(&tout, "Smeti na konci vystupu");
      return 0;
    }

  unsigned val;
  if (!to_uint(&tout, &val))
    tok_err(&tout, "Vystup nezacina cislem");
  unsigned a = get_v(&tout);
  unsigned b = get_v(&tout);
  unsigned c = get_v(&tout);
  unsigned d = get_v(&tout);
  unsigned e = get_v(&tout);
  if (a != e ||
      a == b || a == c || a == d || b == c || b == d || c == d)
    tok_err(&tout, "Vrcholy netvori cyklus");
  unsigned sum = cost(&tout, a, b) + cost(&tout, b, c) + cost(&tout, c, d) + cost(&tout, d, e);
  if (sum != val)
    tok_err(&tout, "Cyklus nema ohlasenou cenu");
  if (get_token(&tout))
    tok_err(&tout, "Smeti na konci vystupu");

  unsigned int opt = get_int(&tcorr);
  if (sum < opt)
    tok_err(&tout, "Nalezeny cyklus neni maximalni");
  if (sum > opt)
    tok_err(&tout, "Bug: Nalezeny cyklus je lepsi nez vzorovy");

  return 0;
}
